#!/usr/bin/env python3

"""
Example usage of the S2I build functions converted from bash to Python.

This demonstrates how to use the s2i_build_as_df_build_args function
that was converted from the container-test-lib.sh bash script.
"""

import sys
import tempfile
from pathlib import Path

from container_ci_suite.container_lib import ContainerTestLib

# Add the container_ci_suite to the path
sys.path.insert(0, str(Path(__file__).parent))


def main():
    """Example usage of S2I build functions."""

    # Initialize the test library
    ct = ContainerTestLib()

    print("=== S2I Build Functions Example ===")

    # Example parameters
    src_image = "registry.access.redhat.com/ubi8/python-38"
    dst_image = "my-python-app:latest"

    # Create a simple test application
    with tempfile.TemporaryDirectory() as app_path:
        app_dir = Path(app_path)

        # Create a simple Python application
        (app_dir / "app.py").write_text("""
from flask import Flask
app = Flask(__name__)

@app.route('/')
def hello():
    return 'Hello from S2I Python app!'

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8080)
""")

        # Create requirements.txt
        (app_dir / "requirements.txt").write_text("Flask==2.0.1\n")

        # Create .s2i directory with environment variables
        s2i_dir = app_dir / ".s2i"
        s2i_dir.mkdir()
        (s2i_dir / "environment").write_text("APP_MODULE=app:app\n")

        print(f"Created test application in: {app_path}")
        print(f"Source image: {src_image}")
        print(f"Destination image: {dst_image}")

        # Example 1: Basic S2I build using s2i_build_as_df
        print("\n1. Testing s2i_build_as_df (wrapper function)...")
        try:
            success = ct.build_as_df(
                app_path=str(app_dir),
                src_image=src_image,
                dst_image=dst_image,
                s2i_args=""
            )
            if success:
                print("✓ S2I build (basic) completed successfully")
            else:
                print("✗ S2I build (basic) failed")
        except Exception as e:
            print(f"✗ S2I build (basic) failed with exception: {e}")

        # Example 2: S2I build with build args using s2i_build_as_df_build_args
        print("\n2. Testing s2i_build_as_df_build_args (full function)...")
        dst_image_with_args = "my-python-app-with-args:latest"
        try:
            success = ct.build_as_df_build_args(
                app_path=str(app_dir),
                src_image=src_image,
                dst_image=dst_image_with_args,
                build_args="--label version=1.0",
                s2i_args="-e FLASK_ENV=development"
            )
            if success:
                print("✓ S2I build (with args) completed successfully")
            else:
                print("✗ S2I build (with args) failed")
        except Exception as e:
            print(f"✗ S2I build (with args) failed with exception: {e}")

        # Example 3: S2I build with mount options
        print("\n3. Testing s2i_build_as_df_build_args with mount options...")
        dst_image_with_mount = "my-python-app-with-mount:latest"
        try:
            success = ct.build_as_df_build_args(
                app_path=str(app_dir),
                src_image=src_image,
                dst_image=dst_image_with_mount,
                build_args="",
                s2i_args="-v /tmp:/tmp:Z -e DEBUG=true"
            )
            if success:
                print("✓ S2I build (with mount) completed successfully")
            else:
                print("✗ S2I build (with mount) failed")
        except Exception as e:
            print(f"✗ S2I build (with mount) failed with exception: {e}")

    print("\n=== S2I Build Examples completed ===")
    print("\nNote: These examples demonstrate the Python conversion of:")
    print("- ct_s2i_build_as_df_build_args() -> s2i_build_as_df_build_args()")
    print("- ct_s2i_build_as_df() -> s2i_build_as_df()")
    print("- ct_get_uid_from_image() -> get_uid_from_image()")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import subprocess
import time
import json

from typing import Any

from container_ci_suite.utils import ContainerTestLibUtils


class PodmanCLIWrapper(object):
    @staticmethod
    def call_podman_command(
        cmd, return_output: bool = True,
        ignore_error: bool = False, shell: bool = True, debug: bool = False, stderr=subprocess.STDOUT
    ):
        """
        Run docker command:
        """
        return ContainerTestLibUtils.run_command(
            f"podman {cmd}",
            return_output=return_output,
            ignore_error=ignore_error,
            shell=shell,
            debug=debug,
            stderr=stderr
        )

    @staticmethod
    def podman_image_exists(image_name: str) -> bool:
        """
        Check if docker image exists or not
        :param image_name: image to check
        :return True: In case if image is present
                False: In case if image is not present
        """
        output = PodmanCLIWrapper.call_podman_command(
            f"images -q {image_name}", ignore_error=True, return_output=True)
        return True if output != "" else False

    @staticmethod
    def podman_inspect(field: str, src_image: str) -> str:
        return PodmanCLIWrapper.call_podman_command(
            f"inspect -f '{field}' {src_image}"
        )

    @staticmethod
    def podman_run_command(cmd):
        return PodmanCLIWrapper.call_podman_command(f"run {cmd}")

    @staticmethod
    def podman_exec_shell_command(
            cid_file_name: str, cmd: str, used_shell: str = "/bin/bash", return_output: bool = True,
            debug: bool = False
    ):
        """
        Function executes shell command if image_name is present in system.
        :param cid_file_name: image to check specified by cid_file_name
        :param cmd: command that will be executed in image
        :param used_shell: which shell will be used /bin/bash or /bin/sh
        :return True: In case if image is present
                False: In case if image is not present
        """
        cmd = f'exec {cid_file_name} {used_shell} -c "{cmd}"'
        print(f"podman exec command is: {cmd}")
        try:
            output = PodmanCLIWrapper.call_podman_command(
                cmd=cmd, return_output=return_output, debug=debug
            )
        except subprocess.CalledProcessError as cpe:
            print(f"podman exec command {cmd} failed. See '{cpe}'")
            return False
        print(f"Output cmd is {output}")
        return output

    @staticmethod
    def podman_run_command_and_remove(
            cid_file_name: str, cmd: str, return_output: bool = True,
            debug: bool = False
    ):
        """
        Function run shell command if image_name is present in system.
        Calling is `podman run --rm {cid_file_name} /bin/bash -c "{cmd}"
        :param cid_file_name: image to check specified by cid_file_name
        :param cmd: command that will be executed in image
        :return True: In case if image is present
                False: In case if image is not present
        """
        cmd = f'run --rm {cid_file_name} /bin/bash -c "{cmd}"'
        print(f"podman exec command is: {cmd}")
        try:
            output = PodmanCLIWrapper.call_podman_command(
                cmd=cmd, return_output=return_output, debug=debug
            )
        except subprocess.CalledProcessError as cpe:
            print(f"podman exec command {cmd} failed. See '{cpe}'")
            return False
        print(f"Output cmd is {output}")
        return output

    @staticmethod
    def podman_get_user_id(src_image, user):
        return PodmanCLIWrapper.call_podman_command(
            f"--rm {src_image} /bin/bash -c 'id -u {user}' 2>/dev/null"
        ).strip()

    @staticmethod
    def podman_pull_image(image_name: str, loops: int = 10) -> bool:
        """
        Function checks if image_name is present in system.
        In case it isn't, try to pull it for specific count of loops
        Default is 10.
        """
        if PodmanCLIWrapper.podman_image_exists(image_name=image_name):
            print("Pulled image already exists.")
            return True
        for loop in range(loops):
            ret_val = PodmanCLIWrapper.call_podman_command(
                cmd=f"pull {image_name}", return_output=False
            )
            if ret_val == 0 and PodmanCLIWrapper.podman_image_exists(image_name=image_name):
                return True
            PodmanCLIWrapper.call_podman_command("images", return_output=True)
            print(f"Pulling of image {image_name} failed. Let's wait {loop*5} seconds and try again.")
            time.sleep(loop*5)
        return False

    @staticmethod
    def podman_inspect_ip_address(container_id: str) -> Any:
        output = PodmanCLIWrapper.call_podman_command(
            f"inspect {container_id}"
        )

        json_output = json.loads(output)
        if len(json_output) == 0:
            return None
        if "NetworkSettings" not in json_output[0]:
            return None
        return json_output[0]["NetworkSettings"]["IPAddress"]

    @staticmethod
    def podman_exit_status(image_name: str) -> str:
        return PodmanCLIWrapper.call_podman_command(
            cmd=f"inspect --format='{{{{.State.ExitCode}}}}' {image_name}",
            return_output=True
        ).strip()

    @staticmethod
    def podman_get_user(image_name: str) -> Any:
        output = PodmanCLIWrapper.call_podman_command(
            f"inspect {image_name}"
        )

        json_output = json.loads(output)
        if len(json_output) == 0:
            return None
        if "Config" not in json_output[0]:
            return None
        return json_output[0]["Config"]["User"]

    @staticmethod
    def podman_get_file_content(cid_file_name: str, filename: str) -> str:
        return PodmanCLIWrapper.call_podman_command(
            cmd=f"exec {cid_file_name} cat {filename}"
        )



#!/bin/env python3
# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import logging
import re
import time
import subprocess
import shutil

from typing import List, Union, Any
from os import getenv
from pathlib import Path
from tempfile import mkdtemp, mktemp

from container_ci_suite.engines.podman_wrapper import PodmanCLIWrapper
from container_ci_suite.utils import ContainerTestLibUtils, cwd
from container_ci_suite.git import Git
from container_ci_suite import utils
from container_ci_suite.exceptions import ContainerCIException

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
logger = logging.getLogger(__name__)


class ContainerImage:

    def __init__(self, image_name: str, cid_file: Path = None, cid_file_dir: Path = None):
        self.image_name: str = image_name
        self.container_args: str = ""
        self.cid_file = cid_file
        self.cid_file_dir = cid_file_dir
        self.app_image_name = "app_dockerfile"
        self.temporary_app_dir: Path = None
        self.application_image_id: str = ""
        logger.info(f"Image name to test: {image_name}")

    def pull_image(self, image_name: str, exit_on_fail: bool = False, loops: int = 10) -> bool:
        """
        Pull an image before test execution.

        Args:
            image_name: Name of the image to pull
            exit_on_fail: Exit if pull fails
            loops: Number of retry attempts

        Returns:
            True if pull successful, False otherwise
        """
        print(f"-> Pulling image {image_name} ...")
        # Check if image is already available locally
        try:
            result = PodmanCLIWrapper.call_podman_command(cmd=f"images -q {image_name}", return_output=True)
            if result.strip():
                print(f"The image {image_name} is already pulled.")
                return True
        except subprocess.CalledProcessError:
            pass

        # Try pulling the image
        for loop in range(1, loops + 1):
            try:
                PodmanCLIWrapper.call_podman_command(cmd=f"pull {image_name}", return_output=False)
                return True
            except subprocess.CalledProcessError:
                print(f"Pulling image {image_name} failed.")
                if loop > loops:
                    print(f"Pulling of image {image_name} failed {loops} times in a row. Giving up.")
                    print(f"!!! ERROR with pulling image {image_name} !!!!")
                    if exit_on_fail:
                        raise ContainerCIException(f"!!! ERROR with pulling image {image_name} !!!!")
                    return False

                wait_time = loop * 5
                print(f"Let's wait {wait_time} seconds and try again.")
                time.sleep(wait_time)

        return False

    @staticmethod
    def is_container_running(container_id: str) -> bool:
        """
        Check if container is in running state.
        Args:
            container_id: Container ID to check
        Returns:
            True if container is running, False otherwise
        """
        try:
            result = PodmanCLIWrapper.call_podman_command(
                cmd=f"inspect -f '{{{{.State.Running}}}}' {container_id}",
                return_output=True
            )
            return result.strip() == "true"
        except subprocess.CalledProcessError:
            return False

    @staticmethod
    def is_container_exists(container_id: str) -> bool:
        """
        Check if container exists.
        Args:
            container_id: Container ID to check
        Returns:
            True if container exists, False otherwise
        """
        try:
            result = PodmanCLIWrapper.call_podman_command(
                cmd=f"ps -q -a -f 'id={container_id}'",
                return_output=True
            )
            return bool(result.strip())
        except subprocess.CalledProcessError:
            return False

    def get_cid(self, cid_file_name: str) -> str:
        """
        Get container ID from cid_file.
        Args:
            cid_file_name: Name of the cid_file
        Returns:
            Container ID
        """
        full_cid_file_name = self.cid_file_dir / cid_file_name
        return utils.get_file_content(full_cid_file_name).strip()

    def get_cip(self, cid_file_name: str) -> str:
        """
        Get container IP address.
        Args:
            cid_file_name: Name of the cid_file
        Returns:
            Container IP address
        """
        container_id = self.get_cid(cid_file_name)
        print(f"Container ID is: {container_id}")
        try:
            result = PodmanCLIWrapper.call_podman_command(
                cmd=f"inspect --format='{{{{.NetworkSettings.IPAddress}}}}' {container_id}",
                return_output=True
            )
            print(f"Container IP is: '{result.strip()}'")
            return result.strip()
        except subprocess.CalledProcessError:
            return ""

    def get_container_exitcode(self, cid_name: str) -> str:
        """
        Get container IP address.
        Args:
            cid_name: Name of the cid_file
        Returns:
            Container IP address
        """
        try:
            container_id = self.get_cid(cid_name)
        except FileNotFoundError:
            return "1"
        print(f"get_container_exit_code: container ID is: {container_id}")
        try:
            result = PodmanCLIWrapper.podman_exit_status(image_name=container_id)
            print(f"get_container_exit_code: '{result.strip()}'.")
            return result
        except subprocess.CalledProcessError:
            return "1"

    @staticmethod
    def wait_for_cid(
        cid_file_name: Union[str, Path],
        max_attempts: int = 10,
        sleep_time: int = 1
    ) -> bool:
        """
        Wait for cid_file to be created.
        Args:
            cid_file_name: Path to cid_file
            max_attempts: Maximum number of attempts
            sleep_time: Sleep time between attempts

        Returns:
            True if cid_file created, False otherwise
        """
        cid_path = Path(cid_file_name)
        for attempt in range(1, max_attempts + 1):
            if cid_path.exists() and cid_path.stat().st_size > 0:
                return True
            print(f"Waiting for container start... {attempt}")
            time.sleep(sleep_time)
        return False

    def rmi_app(self):
        app_cip = self.get_cid_file(Path(self.temporary_app_dir) / self.app_image_name)
        PodmanCLIWrapper.call_podman_command(cmd=f"kill {app_cip}")
        PodmanCLIWrapper.call_podman_command(cmd=f"rmi {self.app_image_name}")
        if Path(self.temporary_app_dir).exists():
            shutil.rmtree(self.temporary_app_dir)

    def s2i_usage(self) -> str:
        return PodmanCLIWrapper.call_podman_command(
            f"run --rm {self.image_name} /bin/bash -c /usr/libexec/s2i/usage"
        )

    def is_image_available(self):
        return PodmanCLIWrapper.call_podman_command(f"inspect {self.image_name}")

    # Replacement for ct_s2i_build_as_df
    def s2i_build_as_df(self, app_path: str, s2i_args: str, src_image: str, dst_image: str):
        named_tmp_dir = mkdtemp()
        tmp_dir = Path(named_tmp_dir)
        if tmp_dir.exists():
            logger.debug("Temporary Directory exists.")
        else:
            logger.debug("Temporary directory not exists.")
        ntf = mktemp(dir=str(tmp_dir), prefix="Dockerfile.")
        df_name = Path(ntf)
        df_content = self.create_dockerfile(
            tmp_dir=tmp_dir,
            app_path=app_path,
            s2i_args=s2i_args,
            src_image=src_image,
            dst_image=dst_image,
        )
        with open(df_name, mode="w") as f:
            f.write('\n'.join(df_content))
        mount_options = utils.get_mount_options_from_s2i_args(s2i_args=s2i_args)
        # Run the build and tag the result
        PodmanCLIWrapper.call_podman_command(
            f"build {mount_options} -f {df_name} --no-cache=true -t {dst_image}"
        )
        return ContainerImage(image_name=dst_image)

    # Replacement for ct_s2i_build_as_df_build_args
    def create_dockerfile(
        self, tmp_dir: Path, app_path: str, s2i_args: str, src_image, dst_image: str
    ) -> Any:
        real_app_path = app_path.replace("file://", "")
        df_content: List = []
        local_scripts: str = "upload/scripts"
        local_app: str = "upload/src"
        os.chdir(tmp_dir)
        if not PodmanCLIWrapper.podman_image_exists(src_image):
            if "pull-policy=never" not in s2i_args:
                PodmanCLIWrapper.call_podman_command(f"pull {src_image}")

        user = PodmanCLIWrapper.podman_inspect(
            field="{{.Config.User}}", src_image=src_image
        )
        if not user:
            user = "0"

        assert int(user)
        user_id = PodmanCLIWrapper.podman_get_user_id(src_image=src_image, user=user)
        if not user_id:
            logger.error(f"id of user {user} not found inside image {src_image}.")
            logger.error("Terminating s2i build.")
            return None
        user_id = user

        incremental: bool = "--incremental" in s2i_args
        if incremental:
            inc_tmp = Path(mktemp(dir=str(tmp_dir), prefix="incremental."))
            ContainerTestLibUtils.run_command(f"setfacl -m 'u:{user_id}:rwx' {inc_tmp}")
            # Check if the image exists, build should fail (for testing use case) if it does not
            if not PodmanCLIWrapper.podman_image_exists(src_image):
                return None
            # Run the original image with a mounted in volume and get the artifacts out of it
            cmd = (
                "if [ -s /usr/libexec/s2i/save-artifacts ];"
                ' then /usr/libexec/s2i/save-artifacts > "$inc_tmp/artifacts.tar";'
                ' else touch "$inc_tmp/artifacts.tar"; fi'
            )
            PodmanCLIWrapper.call_podman_command(
                f"run --rm -v {inc_tmp}:{inc_tmp}:Z {dst_image} bash -c {cmd}"
            )
            # Move the created content into the $tmpdir for the build to pick it up
            shutil.move(f"{inc_tmp}/artifacts.tar", tmp_dir.name)
        real_local_app = tmp_dir / local_app
        real_local_scripts = tmp_dir / local_scripts
        os.makedirs(real_local_app.parent)
        shutil.copytree(real_app_path, real_local_app)
        bin_dir = real_local_app / ".s2i" / "bin"
        if bin_dir.exists():
            shutil.move(bin_dir, real_local_scripts)
        df_content.extend(
            [
                f"FROM {src_image}",
                f"LABEL io.openshift.s2i.build.image={src_image} "
                f"io.openshift.s2i.build.source-location={app_path}",
                "USER root",
                f"COPY {local_app}/ /tmp/src",
            ]
        )
        if real_local_scripts.exists():
            df_content.append(f"COPY {local_scripts} /tmp/scripts")
            df_content.append(f"RUN chown -R {user_id}:0 /tmp/scripts")
        df_content.append(f"RUN chown -R {user_id}:0 /tmp/src")

        # Check for custom environment variables inside .s2i/ folder
        env_file = Path(real_local_app / ".s2i" / "environment")
        if env_file.exists():
            with open(env_file) as fd:
                env_content = fd.readlines()
            # Remove any comments and add the contents as ENV commands to the Dockerfile
            env_content = [f"ENV {x}" for x in env_content if not x.startswith("#")]
            df_content.extend(env_content)

        # Filter out env var definitions from $s2i_args
        # and create Dockerfile ENV commands out of them
        env_content = utils.get_env_commands_from_s2i_args(s2i_args=s2i_args)
        df_content.extend(env_content)

        # Check if CA autority is present on host and add it into Dockerfile
        if utils.get_full_ca_file_path().exists():
            df_content.append(
                "RUN cd /etc/pki/ca-trust/source/anchors && update-ca-trust extract"
            )
        # Add in artifacts if doing an incremental build
        if incremental:
            df_content.extend(
                [
                    "RUN mkdir /tmp/artifacts",
                    "RUN artifacts.tar /tmp/artifacts",
                    f"RUN chown -R {user_id}:0 /tmp/artifacts",
                ]
            )
        df_content.append(f"USER {user_id}")
        # If exists, run the custom assemble script, else default to /usr/libexec/s2i/assemble
        if (real_local_scripts / "assemble").exists():
            df_content.append("RUN /tmp/scripts/assemble")
        else:
            df_content.append("RUN /usr/libexec/s2i/assemble")
        # If exists, set the custom run script as CMD, else default to /usr/libexec/s2i/run
        if Path(real_local_scripts / "run").exists():
            df_content.append("CMD /tmp/scripts/run")
        else:
            df_content.append("CMD /usr/libexec/s2i/run")

        logger.error(df_content)
        return df_content

    def build_image_parse_id(self, dockerfile: str = "", build_params: str = "") -> bool:
        dockerfile_name = f"-f {dockerfile}" if dockerfile != "" else ""
        podman_cmd = f"build --no-cache {dockerfile_name} {build_params}"
        print(f"Command for building container: {podman_cmd}")
        try:
            output = PodmanCLIWrapper.call_podman_command(cmd=podman_cmd, ignore_error=True)
            print(f"Output from build is:\n{output}")
            self.application_image_id = output.split("\n")[-2]
            print(f"Application IMAGE id is: {self.application_image_id}.")
            return True
        except subprocess.CalledProcessError as cpe:
            print(f"Building container by command {podman_cmd} failed for reason '{cpe}' and {cpe.stderr}")
            return False

    def test_response(
            self, url: str = "",
            expected_code: int = 200, port: int = 8080,
            expected_output: str = "", max_tests: int = 20
    ) -> bool:
        url = f"{url}:{port}"
        print(f"URL address to get response from container: {url}")
        cmd_to_run = "curl --connect-timeout 10 -k -s -w '%{http_code}' " + f"{url}"
        # Check if application returns proper HTTP_CODE
        print("Check if HTTP_CODE is valid.")
        for count in range(max_tests):
            try:
                output_code = ContainerTestLibUtils.run_command(cmd=f"{cmd_to_run}", return_output=True)
                return_code = output_code[-3:]
                print(f"Output is: {output_code} and Return Code is: {return_code}")
                try:
                    int_ret_code = int(return_code)
                    if int_ret_code == expected_code:
                        print(f"HTTP_CODE is VALID {int_ret_code}")
                        break
                except ValueError:
                    logger.info(return_code)
                    time.sleep(1)
                    continue
                time.sleep(3)
                continue
            except subprocess.CalledProcessError as cpe:
                print(f"Error from {cmd_to_run} is {cpe.stderr}, {cpe.stdout}")
                time.sleep(3)

        cmd_to_run = "curl --connect-timeout 10 -k -s " + f"{url}"
        # Check if application returns proper output
        for count in range(max_tests):
            output_code = ContainerTestLibUtils.run_command(cmd=f"{cmd_to_run}", return_output=True)
            print(f"Check if expected output {expected_output} is in {cmd_to_run}.")
            if expected_output in output_code:
                print(f"Expected output '{expected_output}' is present.")
                return True
            print(
                f"check_response_inside_cluster:"
                f"expected_output {expected_output} not found in output of {cmd_to_run} command. See {output_code}"
            )
            time.sleep(5)
        return False

    # Replacement for ct_create_container
    def create_container(self, cid_file_name: str, container_args: str = "") -> bool:
        self.cid_file_dir = Path(mkdtemp(suffix=".test_cid_files"))
        p = Path(self.cid_file_dir)
        self.cid_file = p / cid_file_name

        print(f"The CID file {self.cid_file}")
        args_to_run = ""
        if container_args != "":
            args_to_run = container_args
        if container_args != "":
            cmd = f"run --cidfile={self.cid_file} -d {args_to_run} {self.image_name}"
        else:
            cmd = f"run --cidfile={self.cid_file} -d {self.image_name}"
        try:
            PodmanCLIWrapper.call_podman_command(cmd=cmd)
        except subprocess.CalledProcessError as cpe:
            print(f"The command '{cmd}' failed with {cpe.output} and error: {cpe.stderr}")
            print(f"Failed to create container: {cpe}")
            return False
        if not ContainerImage.wait_for_cid(cid_file_name=self.cid_file):
            return False
        print(f"Created container {self.get_cid_file()}")
        return True

    def get_cid_file(self, cid_file: Path = None):
        if cid_file is None:
            return utils.get_file_content(self.cid_file)
        return utils.get_file_content(cid_file)

    # Replacement for ct_check_image_availability
    def check_image_availability(self, public_image_name: str):
        try:
            PodmanCLIWrapper.call_podman_command(
                f"pull {public_image_name}", return_output=False
            )
        except subprocess.CalledProcessError as cfe:
            logger.error(f"{public_image_name} could not be downloaded via 'docker'.")
            logger.error(cfe)
            return False
        return True

    def prepare_app(self, path_to_app: Path) -> bool:
        """
        Prepare test application for S2I build.
        Returns:
            bool: True if preparation successful, False otherwise
        """
        if not self.is_image_available():
            return False
        if not path_to_app.exists():
            return False

        print("ContainerImage(prepare_app): Build the test application image")
        with cwd(path_to_app):
            git_app = Git(path=path_to_app)
            git_app.add_global_config(username="builder", mail="build@localhost")
            git_app.add_files()
            git_app.commit_files(message="Sample commit")
        return True

    # Replacement for ct_clean_containers
    def cleanup_container(self):
        logger.info(f"Cleaning CID_FILE_DIR {self.cid_file_dir} is ongoing.")
        p = Path(self.cid_file_dir)
        cid_files = p.glob("*")
        for cid_file in cid_files:
            if not Path(cid_file).exists():
                continue
            container_id = utils.get_file_content(cid_file)
            logger.info("Stopping container")
            PodmanCLIWrapper.call_podman_command(f"stop {container_id}")
            exit_code = PodmanCLIWrapper.podman_inspect(
                field="{{.State.ExitCode}}", src_image=container_id
            )
            if exit_code != 0:
                logs = PodmanCLIWrapper.call_podman_command(f"logs {container_id}")
                logger.info(logs)
            PodmanCLIWrapper.call_podman_command(f"rm -v {container_id}")
            if not Path(cid_file).exists():
                continue
            cid_file.unlink()
        os.rmdir(self.cid_file_dir)
        logger.info(f"Cleaning CID_FILE_DIR {self.cid_file_dir} is DONE.")

    # Replacement for ct_assert_container_creation_fails
    def assert_container_fails(self, cid_file: str, container_args: str):
        attempt: int = 1
        max_attempts: int = 10
        old_container_args = container_args
        if self.create_container(cid_file, container_args=container_args):
            cid = self.get_cid_file()
            while not PodmanCLIWrapper.podman_inspect(
                field="{{.State.Running}}", src_image=cid
            ):
                time.sleep(2)
                attempt += 1
                if attempt > max_attempts:
                    PodmanCLIWrapper.call_podman_command("stop cid")
                    return True
            exit_code = PodmanCLIWrapper.podman_inspect(
                field="{{.State.ExitCode}}", src_image=cid
            )
            if exit_code == 0:
                return True
            PodmanCLIWrapper.call_podman_command(f"rm -v {cid}")
            self.cid_file.unlink()
        if old_container_args != "":
            self.container_args = old_container_args
        return False

    # Replacement for ct_npm_works
    def npm_works(self):
        tempdir = mkdtemp(suffix="npm_test")
        self.cid_file = Path(tempdir) / "npm_test_cid"
        try:
            try:
                PodmanCLIWrapper.call_podman_command(
                    f'run --rm {self.image_name} /bin/bash -c "npm --version"'
                )
            except subprocess.CalledProcessError:
                logger.error(
                    f"ERROR: 'npm --version' does not work inside the image {self.image_name}."
                )
                return False

            # TODO
            # Add {self.image_name}-testapp as soon as function `s2i_create_df` is ready.
            PodmanCLIWrapper.call_podman_command(
                f"run -d {utils.get_mount_ca_file()} --rm --cidfile={self.cid_file} {self.image_name}"
            )
            if not self.wait_for_cid(self.cid_file):
                logger.error("Container did not create cidfile.")
                return False

            try:
                jquery_output = PodmanCLIWrapper.call_podman_command(
                    f"exec {self.get_cid_file(self.cid_file)} "
                    f"/bin/bash -c "
                    f"'npm --verbose install jquery && test -f node_modules/jquery/src/jquery.js'"
                )
            except subprocess.CalledProcessError:
                logger.error(
                    f"npm could not install jquery inside the image ${self.image_name}."
                )
                return False
            if getenv("NPM_REGISTRY") and utils.get_full_ca_file_path().exists():
                if utils.get_os_environment("NPM_REGISTRY") in jquery_output:
                    logger.error("Internal repository is NOT set. Even it is requested.")
                    return False

            if self.cid_file.exists():
                PodmanCLIWrapper.call_podman_command(
                    f"stop {self.get_cid_file(self.cid_file)}"
                )
                self.cid_file.unlink()
            print("Success!")
            logger.info("Npm works.")
            return True
        finally:
            shutil.rmtree(tempdir)

    # Replacement for ct_binary_found_from_df
    def binary_found_from_df(self, binary: str = "", binary_path: str = "^/opt/rh"):
        tempdir = mkdtemp(suffix=f"{self.image_name}_binary")
        dockerfile = Path(tempdir) / "Dockerfile"
        logger.info(f"Testing {binary} in build from Dockerfile")
        content: str = f"""FROM {self.image_name}
RUN which {binary} | grep {binary_path}
        """
        with open(dockerfile, "w") as f:
            f.write(content)
        if not PodmanCLIWrapper.call_podman_command(
            f"build -f {dockerfile} --no-cache {tempdir}", return_output=False
        ):
            logger.error(f"Failed to find {binary} in Dockerfile!")
            return False
        return True

    def doc_content_old(self, strings: List) -> bool:
        logger.info("Testing documentation in the container image")
        files_to_check = ["help.1"]
        for f in files_to_check:
            doc_content = PodmanCLIWrapper.podman_run_command(f'--rm {self.image_name} /bin/bash -c cat {f}')
            for term in strings:
                # test = re.search(f"{term}", doc_content)
                logger.info(f"ERROR: File /{f} does not contain '{term}'.")
                return False
            for term in ["TH", "PP", "SH"]:
                if term not in doc_content:
                    logger.info(f"ERROR: help.1 is probably not in troff or groff format, since {term} is missing")
                    return False
        return True

    # Replacement for ct_check_exec_env_vars
    def test_check_exec_env_vars(self, env_filter: str = "^X_SCLS=|/opt/rh|/opt/app-root"):
        check_envs = PodmanCLIWrapper.podman_run_command(f'--rm {self.image_name} /bin/bash -c env')
        logger.debug(f"Run envs {check_envs}")
        self.create_container(cid_file_name="exec_env_vars", container_args="bash -c 'sleep 1000'")
        loop_envs = PodmanCLIWrapper.call_podman_command(f"exec {self.get_cid_file(self.cid_file)} env")
        self.test_check_envs_set(env_filter=env_filter, check_envs=check_envs, loop_envs=loop_envs)

    # Replacement for ct_check_scl_enable_vars
    def test_check_envs_set(self, env_filter: str, check_envs: str, loop_envs: str, env_format="VALUE"):
        fields_to_check: List = [
            x for x in loop_envs.split('\n') if re.findall(env_filter, x) and not x.startswith("PWD=")
        ]
        for field in fields_to_check:
            var_name, stripped = field.split('=', 2)
            filtered_envs = [x for x in check_envs.split('\n') if x.startswith(f"{var_name}=")]
            if not filtered_envs:
                logger.error(f"{var_name} not found during 'docker exec'")
                return False
            filter_envs = ''.join(filtered_envs)
            for value in stripped.split(':'):
                # If the value checked does not go through env_filter we do not care about it
                ret = re.findall(env_filter, value)
                if not ret:
                    continue
                new_env = env_format.replace('VALUE', value)
                find_env = re.findall(rf"{new_env}", filter_envs)
                if not find_env:
                    logger.error(f"Value {value} is missing from variable {var_name}")
                    logger.error(filtered_envs)
                    return False
        return True

    def build_test_container(
            self,
            dockerfile: str, app_url: str, app_dir: str,
            build_args: str = "",
            app_image_name: str = "app_dockerfile"
    ) -> bool:
        """
        Function builds testing application container
        param: dockerfile - path to a Dockerfile that will be used for building an image
                         (must work with an application directory called 'app-src')
        param: app_url - git or local URI with a testing application, supports "@" to indicate a different branch
        param: expected_output - PCRE regular expression that must match the response body
        param: app_dir - name of the application directory that is used in the Dockerfile
        param: build_args - build args that will be used for building an image
        param: app_image_name - how the container will be named in output `podman images`. Default app_dockerfile
        """
        self.temporary_app_dir = Path(mkdtemp(suffix="app_test_dir"))
        self.app_image_name = app_image_name
        if not app_dir:
            print("build_test_container: Parameter app_dir has to be set.")
            return False
        if not Path(dockerfile).exists():
            print(f"build_test_container: Dockerfile {dockerfile} does not exist or is empty.")
            return False
        full_path = Path(dockerfile).resolve()
        print(f"Resolved dockerfile {full_path}")
        tempdir = self.temporary_app_dir
        with utils.cwd(tempdir) as _:
            print(f"Copy Dockerfile from {full_path} to '{tempdir}/Dockerfile'")
            shutil.copy(full_path, "Dockerfile")
            podman_content = utils.get_file_content(Path("Dockerfile")).split('\n')
            for index, line in enumerate(podman_content):
                podman_content[index] = re.sub("^FROM.*$", f"FROM  {self.image_name}", line)
            utils.save_file_content('\n'.join(podman_content), Path("Dockerfile"))
            if Path(app_url).is_dir():
                print(f"Copy local folder {app_url} to {app_dir}.")
                shutil.copytree(app_url, app_dir, symlinks=True)
            else:
                utils.clone_git_repository(app_url=app_url, app_dir=app_dir)
            print(f"Building '{app_image_name}' image using docker build")
            if not self.build_image_parse_id(build_params=f"-t {self.app_image_name} . {build_args}"):
                return False
        output = PodmanCLIWrapper.call_podman_command(cmd="images", ignore_error=True)
        print(f"Output from podman images is:\n{output}")
        return True

    def test_app_dockerfile(self) -> bool:
        """
            param: expected_output - PCRE regular expression that must match the response body
        """
        podman_cmd = f"run -d --cidfile={self.cid_file_dir}/{self.app_image_name} --rm {self.app_image_name}"
        print(f"Run container {self.app_image_name}: {podman_cmd}")
        try:
            output = PodmanCLIWrapper.call_podman_command(cmd=podman_cmd, ignore_error=True).strip()
            print(f"Output from {podman_cmd} is:\n'{output}'.")
        except subprocess.CalledProcessError as cpe:
            print(f"Building container by command {podman_cmd} failed for reason '{cpe}' and '{cpe.stderr}'")
            return False
        if not ContainerImage.wait_for_cid(
            cid_file_name=f"{self.cid_file_dir}/{self.app_image_name}"
        ):
            print("Container did not create cidfile. See logs from container.")
            return False
        return True

    def get_logs(self, cid_file_name: str):
        container_id = self.get_cid(cid_file_name=cid_file_name)
        return PodmanCLIWrapper.call_podman_command(cmd=f"logs {container_id}", return_output=True, ignore_error=True)

    def get_logs_std_output(self, expected: str, cid_file_name: str, std_output: str = None) -> bool:
        container_id = self.get_cid(cid_file_name=cid_file_name)
        result = ""
        if std_output == "stdouterr" or std_output is None:
            result = PodmanCLIWrapper.call_podman_command(cmd=f"logs {container_id}", return_output=True)
        if std_output == "stdout":
            result = PodmanCLIWrapper.call_podman_command(
                cmd=f"logs {container_id}", return_output=True,
                stderr=subprocess.DEVNULL
            )
        if std_output == "stderr":
            result = PodmanCLIWrapper.call_podman_command(
                cmd=f"logs {container_id}",
                stderr=subprocess.PIPE
            )
        if not re.search(expected, result):
            print(f'ERROR[docker logs {std_output or ""}] Expected \'{expected}\', got \'{result}\'')
            return False

        return True

    def test_command(self, run_cmd: str, expected: str) -> bool:
        """
        Test command execution in container using both bash and sh.
        Args:
            run_cmd (str): Command to execute in container
            expected (str): Expected output pattern
        Returns:
            bool: True if test passes, False otherwise
        """
        print(f"Test command ({run_cmd})")
        result = PodmanCLIWrapper.podman_exec_shell_command(cid_file_name=self.image_name, cmd=run_cmd)
        if not re.search(expected, result):
            print(f'ERROR[exec /bin/bash -c "{run_cmd}"] Expected \'{expected}\', got \'{result}\'')
            return False

        result = PodmanCLIWrapper.podman_exec_shell_command(
            cid_file_name=self.image_name, used_shell="/usr/sh", cmd=run_cmd
        )
        if not re.search(expected, result):
            print(f'ERROR[exec /bin/sh -c "{run_cmd}"] Expected \'{expected}\', got \'{result}\'')
            return False

        return True



#!/usr/bin/env python3
"""
Comprehensive pytest suite for the DockerfileProcessor module.
This test suite covers all functionality of the dockerfile_processor module
including the replacement of sed commands with Python equivalents.
"""

import os
import sys
import pytest
import tempfile
import shutil
from pathlib import Path

from container_ci_suite.dockerfile_processor import DockerfileProcessor


# Add the test directory to the Python path
TEST_DIR = Path(os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, str(TEST_DIR))


@pytest.mark.unit
@pytest.mark.dockerfile
class TestDockerfileProcessor:
    """Test suite for DockerfileProcessor class."""

    @pytest.fixture
    def sample_dockerfile_content(self):
        """Sample Dockerfile content for testing."""
        return """FROM registry.redhat.io/ubi8/ubi:latest

LABEL maintainer="Test Team <test@example.com>"

ENV NGINX_VERSION=1.20 \
    NGINX_SHORT_VER=120 \
    NAME=nginx \
    VERSION=0

# Install nginx
RUN yum install -y nginx-$NGINX_VERSION && \
    yum clean all

EXPOSE 8080

COPY ./s2i/bin/ $STI_SCRIPTS_PATH

USER 1001

CMD ["nginx", "-g", "daemon off;"]
"""

    @pytest.fixture
    def temp_dockerfile(self, sample_dockerfile_content):
        """Create a temporary Dockerfile for testing."""
        fd, temp_path = tempfile.mkstemp(suffix='.dockerfile', prefix='test_')
        try:
            with os.fdopen(fd, 'w') as f:
                f.write(sample_dockerfile_content)
            yield temp_path
        finally:
            if os.path.exists(temp_path):
                os.unlink(temp_path)

    @pytest.fixture
    def temp_dir(self):
        """Create a temporary directory for testing."""
        temp_dir = tempfile.mkdtemp(prefix='dockerfile_test_')
        try:
            yield temp_dir
        finally:
            shutil.rmtree(temp_dir, ignore_errors=True)

    def test_dockerfile_processor_init_valid_file(self, temp_dockerfile):
        """Test DockerfileProcessor initialization with valid file."""
        processor = DockerfileProcessor(temp_dockerfile)
        assert processor.dockerfile_path == Path(temp_dockerfile)
        assert processor.dockerfile_path.exists()

    def test_dockerfile_processor_init_invalid_file(self):
        """Test DockerfileProcessor initialization with invalid file."""
        with pytest.raises(FileNotFoundError):
            DockerfileProcessor("/nonexistent/dockerfile")

    def test_process_nginx_version_env_replacement(self, temp_dockerfile):
        """Test replacement of ENV NGINX_VERSION line."""
        processor = DockerfileProcessor(temp_dockerfile)
        processor.update_env_in_dockerfile(version="1.24", what_to_replace="ENV NGINX_VERSION")

        # Check that ENV NGINX_VERSION was replaced
        assert "ENV NGINX_VERSION=1.24" in processor.get_content()
        assert "ENV NGINX_VERSION=1.20" not in processor.get_content()

    def test_process_nginx_version_variable_replacement(self, temp_dockerfile):
        """Test replacement of $NGINX_VERSION variables."""
        processor = DockerfileProcessor(temp_dockerfile)
        processor.update_variable_in_dockerfile(version="1.24", variable="NGINX_VERSION")

        # Check that $NGINX_VERSION was replaced
        assert "nginx-1.24" in processor.get_content()
        assert "nginx-$NGINX_VERSION" not in processor.get_content()

    def test_process_nginx_version_with_output_file(self, temp_dockerfile, temp_dir):
        """Test processing with output file."""
        processor = DockerfileProcessor(temp_dockerfile)
        output_path = Path(temp_dir) / "output.dockerfile"

        processor.update_env_in_dockerfile(version="1.26", what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version="1.26", variable="NGINX_VERSION")
        output_path = Path(processor.create_temp_dockerfile())
        # Check that output file was created
        assert output_path.exists()

        # Check content of output file
        with open(output_path, 'r') as f:
            content = f.read()

        assert "ENV NGINX_VERSION=1.26" in content
        assert "nginx-1.26" in content

    def test_create_temp_dockerfile(self, temp_dockerfile):
        """Test creation of temporary Dockerfile."""
        processor = DockerfileProcessor(temp_dockerfile)
        processor.update_env_in_dockerfile(version="1.22", what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version="1.22", variable="NGINX_VERSION")

        temp_path = processor.create_temp_dockerfile()

        try:
            assert os.path.exists(temp_path)

            with open(temp_path, 'r') as f:
                content = f.read()

            assert "ENV NGINX_VERSION=1.22" in content
            assert "nginx-1.22" in content
        finally:
            if os.path.exists(temp_path):
                os.unlink(temp_path)

    @pytest.mark.parametrize("version", [
        "1.20",
        "1.22",
        "1.24",
        "1.26",
        "1.22-micro",
        "1.24-micro"
    ])
    def test_process_different_versions(self, temp_dockerfile, version):
        """Test processing with different version formats."""
        processor = DockerfileProcessor(temp_dockerfile)
        processor.update_env_in_dockerfile(version=version, what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version=version, variable="NGINX_VERSION")

        expected_env = f"ENV NGINX_VERSION={version}"
        expected_var = f"nginx-{version}"

        assert expected_env in processor.get_content()
        assert expected_var in processor.get_content()

    def test_validate_dockerfile_syntax_valid(self):
        """Test validation with valid Dockerfile content."""
        processor = DockerfileProcessor.__new__(DockerfileProcessor)  # Create without calling __init__

        valid_content = """FROM ubuntu:20.04
RUN apt-get update
ENV TEST=value
EXPOSE 80
CMD ["nginx"]
"""
        assert processor.validate_dockerfile_syntax(valid_content) is True

    def test_validate_dockerfile_syntax_invalid_no_from(self):
        """Test validation with invalid Dockerfile (no FROM)."""
        processor = DockerfileProcessor.__new__(DockerfileProcessor)

        invalid_content = """RUN apt-get update
ENV TEST=value
"""
        assert processor.validate_dockerfile_syntax(invalid_content) is False

    def test_validate_dockerfile_syntax_invalid_instruction(self):
        """Test validation with invalid instruction."""
        processor = DockerfileProcessor.__new__(DockerfileProcessor)

        invalid_content = """FROM ubuntu:20.04
INVALID_INSTRUCTION something
"""
        assert processor.validate_dockerfile_syntax(invalid_content) is False

    def test_validate_dockerfile_syntax_with_comments(self):
        """Test validation with comments and empty lines."""
        processor = DockerfileProcessor.__new__(DockerfileProcessor)

        content_with_comments = """# This is a comment
FROM ubuntu:20.04

# Another comment
RUN apt-get update
ENV TEST=value

# Final comment
CMD ["nginx"]
"""
        assert processor.validate_dockerfile_syntax(content_with_comments) is True

    def test_process_preserves_other_content(self, temp_dockerfile):
        """Test that processing preserves other Dockerfile content."""
        processor = DockerfileProcessor(temp_dockerfile)
        processor.update_env_in_dockerfile(version="1.24", what_to_replace="ENV NGINX_VERSION")

        # Check that other content is preserved
        assert "FROM registry.redhat.io/ubi8/ubi:latest" in processor.content
        assert "LABEL maintainer" in processor.content
        assert "EXPOSE 8080" in processor.content
        assert "USER 1001" in processor.content
        assert "CMD [\"nginx\"" in processor.content

    def test_multiple_env_nginx_version_lines(self, temp_dir):
        """Test handling of multiple ENV NGINX_VERSION lines."""
        content = """FROM ubuntu:20.04
ENV NGINX_VERSION=1.20
RUN some command
ENV NGINX_VERSION=1.22
CMD ["nginx"]
"""

        dockerfile_path = Path(temp_dir) / "multi_env.dockerfile"
        with open(dockerfile_path, 'w') as f:
            f.write(content)

        processor = DockerfileProcessor(dockerfile_path)
        processor.update_env_in_dockerfile(version="1.26", what_to_replace="ENV NGINX_VERSION")

        # Both lines should be replaced
        assert processor.get_content().count("ENV NGINX_VERSION=1.26") == 2
        assert "ENV NGINX_VERSION=1.20" not in processor.get_content()
        assert "ENV NGINX_VERSION=1.22" not in processor.get_content()

    def test_no_nginx_version_lines(self, temp_dir):
        """Test handling when no NGINX_VERSION lines exist."""
        content = """FROM ubuntu:20.04
ENV OTHER_VAR=value
RUN some command
CMD ["nginx"]
"""

        dockerfile_path = Path(temp_dir) / "no_nginx.dockerfile"
        with open(dockerfile_path, 'w') as f:
            f.write(content)

        processor = DockerfileProcessor(dockerfile_path)
        processor.update_env_in_dockerfile(version="1.24", what_to_replace="ENV NGINX_VERSION")

        # Content should be unchanged except no $NGINX_VERSION to replace
        assert "ENV OTHER_VAR=value" in processor.get_content()
        assert "ENV NGINX_VERSION" not in processor.get_content()

    def test_edge_case_empty_file(self, temp_dir):
        """Test handling of empty Dockerfile."""
        dockerfile_path = Path(temp_dir) / "empty.dockerfile"
        with open(dockerfile_path, 'w') as f:
            f.write("")

        processor = DockerfileProcessor(dockerfile_path)
        processor.update_env_in_dockerfile(version="1.26", what_to_replace="ENV NGINX_VERSION")

        assert processor.content == ""

    def test_edge_case_version_in_comments(self, temp_dir):
        """Test that versions in comments are not replaced."""
        content = """FROM ubuntu:20.04
# This uses NGINX_VERSION=1.20 but should not be replaced
ENV NGINX_VERSION=1.22
# Another comment with $NGINX_VERSION
RUN nginx-$NGINX_VERSION
"""

        dockerfile_path = Path(temp_dir) / "comments.dockerfile"
        with open(dockerfile_path, 'w') as f:
            f.write(content)

        processor = DockerfileProcessor(dockerfile_path)
        processor.update_env_in_dockerfile(version="1.26", what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version="1.26", variable="NGINX_VERSION")

        # ENV line should be replaced
        assert "ENV NGINX_VERSION=1.26" in processor.get_content()
        # Variable should be replaced
        assert "nginx-1.26" in processor.get_content()
        # Comments should remain unchanged
        assert "# This uses NGINX_VERSION=1.20 but should not be replaced" in processor.get_content()
        assert "# Another comment with $NGINX_VERSION" in processor.get_content()


@pytest.mark.integration
@pytest.mark.dockerfile
class TestDockerfileProcessorIntegration:
    """Integration tests that simulate the actual use case from the shell script."""

    @pytest.fixture
    def example_dockerfile_content(self):
        """Content similar to the actual examples/Dockerfile."""
        return """FROM registry.redhat.io/ubi8/s2i-base:1

ENV NGINX_VERSION=1.20 \
    NGINX_SHORT_VER=120 \
    NAME=nginx \
    VERSION=0

LABEL summary="Platform for running nginx or building nginx-based application" \
      description="Nginx $NGINX_VERSION available as container is a web server and a reverse proxy server."

RUN yum install -y centos-release-scl && \
    INSTALL_PKGS="nginx116 nginx116-nginx nginx116-nginx-mod-stream" && \
    yum install -y --setopt=tsflags=nodocs $INSTALL_PKGS && \
    yum clean all

COPY ./s2i/bin/ $STI_SCRIPTS_PATH
COPY ./root/ /

RUN /usr/libexec/s2i/assemble

USER 1001

EXPOSE 8080

CMD $STI_SCRIPTS_PATH/run
"""

    @pytest.fixture
    def example_dockerfile(self, example_dockerfile_content, temp_dir):
        """Create example Dockerfile for integration testing."""
        dockerfile_path = Path(temp_dir) / "Dockerfile"
        with open(dockerfile_path, 'w') as f:
            f.write(example_dockerfile_content)
        return dockerfile_path

    def test_integration_sed_replacement_equivalent(self, example_dockerfile, temp_dir):
        """Test that our Python replacement produces equivalent results to sed command."""
        # Test with version that would be used in actual script
        test_version = "1.24"

        processor = DockerfileProcessor(example_dockerfile)
        processor.update_env_in_dockerfile(version=test_version, what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version=test_version, variable="NGINX_VERSION")

        # Verify the exact transformations that sed would do
        assert f"ENV NGINX_VERSION={test_version}" in processor.get_content()
        assert "ENV NGINX_VERSION=1.20" not in processor.get_content()
        assert f"Nginx {test_version} available" in processor.get_content()
        assert "Nginx $NGINX_VERSION available" not in processor.get_content()

    def test_integration_micro_version(self, example_dockerfile):
        """Test with micro version as used in the actual script."""
        test_version = "1.22-micro"

        processor = DockerfileProcessor(example_dockerfile)
        processor.update_env_in_dockerfile(version="1.22-micro", what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version="1.22-micro", variable="NGINX_VERSION")

        assert f"ENV NGINX_VERSION={test_version}" in processor.get_content()
        assert f"Nginx {test_version} available" in processor.get_content()

    def test_integration_dockerfile_s2i(self, temp_dir):
        """Test with Dockerfile.s2i content."""
        s2i_content = """FROM registry.redhat.io/ubi8/s2i-base:1

ENV NGINX_VERSION=1.20

LABEL io.k8s.description="Nginx $NGINX_VERSION" \
      io.k8s.display-name="Nginx $NGINX_VERSION"

USER 1001
"""

        dockerfile_path = Path(temp_dir) / "Dockerfile.s2i"
        with open(dockerfile_path, 'w') as f:
            f.write(s2i_content)

        processor = DockerfileProcessor(dockerfile_path)
        processor.update_env_in_dockerfile(version="1.26", what_to_replace="ENV NGINX_VERSION")
        processor.update_variable_in_dockerfile(version="1.26", variable="NGINX_VERSION")

        assert "ENV NGINX_VERSION=1.26" in processor.get_content()
        assert "Nginx 1.26" in processor.get_content()
        # Should appear twice in the labels
        assert processor.get_content().count("1.26") == 3  # ENV + 2 labels



"""
Integration tests for ContainerTestLib class.
These tests require Docker to be available and may be slower.
"""

import pytest

from unittest.mock import patch
from subprocess import CalledProcessError

from container_ci_suite.container_lib import ContainerTestLib
from container_ci_suite.engines.container import ContainerImage


@pytest.mark.integration
class TestContainerIntegration:
    """Integration tests for container operations."""
    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_create_container_success(
            self, mock_docker_commands, mock_file_operations
    ):
        """Test successful container creation."""
        self.lib.image_name = "test:latest"

        # Mock the CID file creation
        cid_file_path = str(self.lib.cid_file_dir / "test_container")
        mock_file_operations[cid_file_path] = "container123"

        with patch.object(ContainerImage, 'wait_for_cid', return_value=True):
            result = self.lib.create_container("test_container", "sleep 60")
            assert result is True

    def test_ct_create_container_failure(self):
        """Test failed container creation."""
        self.lib.image_name = "test:latest"

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.side_effect = CalledProcessError(1, "docker run")
            result = self.lib.create_container("test_container", "sleep 60")
            assert result is False

    def test_ct_assert_container_creation_fails_success(
            self, mock_docker_commands, mock_file_operations
    ):
        """Test assertion that container creation should fail."""

        # Mock container that exits with non-zero status
        with patch.object(self.lib, 'create_container', return_value=True), \
             patch.object(self.lib, 'get_cid', return_value="container123"), \
             patch.object(ContainerImage, 'is_container_running', return_value=False), \
             patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:

            # Mock exit status check
            mock_cmd.return_value = "1"  # Non-zero exit code

            result = self.lib.assert_container_creation_fails("--invalid-arg")
            assert result is True


@pytest.mark.integration
class TestImageIntegration:
    """Integration tests for image operations."""
    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_image_build_and_cleanup(
            self, temp_dir, sample_dockerfile_content
    ):
        """Test building image and cleanup."""

        # Create a test Dockerfile
        dockerfile = temp_dir / "Dockerfile"
        dockerfile.write_text(sample_dockerfile_content)

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = "Successfully built abc123\nabc123"

            result = self.lib.build_image_and_parse_id(str(dockerfile), str(temp_dir))
            assert result is True
            assert self.lib.app_image_id == "abc123"

    def test_binary_found_from_dockerfile(self):
        """Test binary availability check from Dockerfile."""
        self.lib.image_name = "test:latest"

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = "Successfully built def456\ndef456"

            result = self.lib.binary_found_from_df("node", "/usr/bin")
            assert result is True


@pytest.mark.integration
class TestS2IIntegration:
    """Integration tests for S2I operations."""

    def test_s2i_usage(self, mock_docker_commands):
        """Test S2I usage command."""
        lib = ContainerTestLib(image_name="test:latest")
        result = lib.s2i_usage()
        assert result == ""  # Mocked to return empty string


@pytest.mark.integration
class TestEnvironmentIntegration:
    """Integration tests for environment checking."""

    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_check_exec_env_vars_success(
            self, sample_environment_vars
                                            ):
        """Test environment variable checking between run and exec."""
        self.lib.image_name = "test:latest"

        env_output = "\n".join([f"{k}={v}" for k, v in sample_environment_vars.items()])

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd, \
             patch.object(self.lib, 'create_container', return_value=True), \
             patch.object(self.lib, 'get_cid', return_value="container123"):

            mock_cmd.return_value = env_output

            result = self.lib.check_exec_env_vars()
            assert result is True

    def test_ct_check_scl_enable_vars_success(
            self, sample_environment_vars
    ):
        """Test SCL environment variable checking."""
        self.lib.image_name = "test:latest"

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            # Mock X_SCLS output
            mock_cmd.side_effect = [
                "nodejs16",  # X_SCLS value
                "\n".join([f"{k}={v}" for k, v in sample_environment_vars.items()]),  # env output
                "\n".join([f"{k}={v}{v}" for k, v in sample_environment_vars.items()])  # scl enable env
            ]

            result = self.lib.check_scl_enable_vars()
            assert result is True


@pytest.mark.integration
@pytest.mark.network
class TestNetworkIntegration:
    """Integration tests requiring network access."""
    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_test_response_success(self):
        """Test HTTP response testing."""
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = "Welcome to the app200"

            result = self.lib.test_response(
                "http://localhost:8080",
                200,
                "Welcome",
                max_attempts=1
            )
            assert result is True

    def test_test_response_failure(self):
        """Test HTTP response testing with failure."""
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = "Error404"

            result = self.lib.test_response(
                "http://localhost:8080",
                200,
                "Welcome",
                max_attempts=1
            )
            assert result is False


@pytest.mark.integration
class TestNpmIntegration:
    """Integration tests for NPM functionality."""
    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_npm_works_success(self, temp_dir):
        """Test NPM functionality check."""
        self.lib.image_name = "test:latest"

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd, \
                patch.object(ContainerImage, 'wait_for_cid', return_value=True), \
                patch('container_ci_suite.utils.get_file_content', return_value="container123"):
            mock_cmd.side_effect = [
                "8.19.2",  # npm --version
                "",  # docker run for test container
                "npm install success",  # npm install jquery
                ""  # docker stop
            ]

            result = self.lib.npm_works()
            assert result is True

    def test_ct_npm_works_version_failure(self):
        """Test NPM functionality check with version failure."""
        self.lib.image_name = "test:latest"

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.side_effect = CalledProcessError(1, "npm --version")

            result = self.lib.npm_works()
            assert result is False


@pytest.mark.integration
class TestDocumentationIntegration:
    """Integration tests for documentation checking."""

    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_doc_content_old_success(self):
        """Test documentation content checking."""
        doc_content = """.TH NODEJS 1 "nodejs container"
.PP
This is a nodejs container
.SH DESCRIPTION
Node.js runtime for applications
"""

        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = doc_content

            result = self.lib.doc_content_old(["nodejs", "container"])
            assert result is True

    def test_doc_content_old_missing_string(self):
        """Test documentation content checking with missing string."""
        doc_content = """.TH NODEJS 1 "nodejs container"
.PP
This is a nodejs container
"""
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = doc_content

            result = self.lib.doc_content_old(["missing_string"])
            assert result is False

    def test_doc_content_old_invalid_format(self):
        """Test documentation content checking with invalid format."""
        doc_content = "This is not a proper troff format"
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = doc_content

            result = self.lib.doc_content_old(["nodejs"])
            assert result is False


@pytest.mark.integration
class TestCertificateIntegration:
    """Integration tests for certificate operations."""
    def setup_method(self):
        self.lib = ContainerTestLib()

    def test_gen_self_signed_cert_pem(self, temp_dir):
        """Test self-signed certificate generation."""
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.return_value = ""  # Successful command execution

            result = self.lib.gen_self_signed_cert_pem(str(temp_dir), "test")
            assert result is True

            # Check that openssl commands were called
            assert mock_cmd.call_count >= 2

    def test_gen_self_signed_cert_pem_failure(self, temp_dir):
        """Test self-signed certificate generation failure."""
        with patch('container_ci_suite.utils.ContainerTestLibUtils.run_command') as mock_cmd:
            mock_cmd.side_effect = CalledProcessError(1, "openssl")

            result = self.lib.gen_self_signed_cert_pem(str(temp_dir), "test")
            assert result is False



#!/usr/bin/env python3

"""
Test script for the s2i_multistage_build function migration.
This script demonstrates the usage of the migrated ct_s2i_multistage_build function.
"""

import tempfile
import shutil
from pathlib import Path
from container_ci_suite.container_lib import ContainerTestLib


def test_s2i_multistage_build():
    """Test the s2i_multistage_build function with a simple example."""

    # Initialize the container test library
    ctl = ContainerTestLib()

    # Create a temporary directory with a simple app
    test_app_dir = Path(tempfile.mkdtemp(prefix="test_app_"))

    try:
        # Create a simple test application
        (test_app_dir / "index.html").write_text("""
<!DOCTYPE html>
<html>
<head>
    <title>Test App</title>
</head>
<body>
    <h1>Hello from S2I Multistage Build!</h1>
</body>
</html>
""")

        # Create a simple package.json for Node.js app
        (test_app_dir / "package.json").write_text("""
{
  "name": "test-app",
  "version": "1.0.0",
  "description": "Test application for S2I multistage build",
  "main": "server.js",
  "scripts": {
    "start": "node server.js"
  }
}
""")

        # Create a simple server.js
        (test_app_dir / "server.js").write_text("""
const http = require('http');
const fs = require('fs');

const server = http.createServer((req, res) => {
    if (req.url === '/') {
        fs.readFile('index.html', (err, data) => {
            if (err) {
                res.writeHead(404);
                res.end('Not found');
                return;
            }
            res.writeHead(200, {'Content-Type': 'text/html'});
            res.end(data);
        });
    } else {
        res.writeHead(404);
        res.end('Not found');
    }
});

const port = process.env.PORT || 8080;
server.listen(port, () => {
    console.log(`Server running on port ${port}`);
});
""")

        print("Testing s2i_multistage_build function...")
        print(f"Test app directory: {test_app_dir}")

        # Test parameters (using example images - these would need to be real images in actual use)
        app_path = str(test_app_dir)
        src_image = "registry.access.redhat.com/ubi8/nodejs-16"  # Builder image
        sec_image = "registry.access.redhat.com/ubi8/nodejs-16-minimal"  # Runtime image
        dst_image = "test-multistage-app:latest"
        s2i_args = "-e NODE_ENV=production"

        print(f"App path: {app_path}")
        print(f"Source image: {src_image}")
        print(f"Second stage image: {sec_image}")
        print(f"Destination image: {dst_image}")
        print(f"S2I args: {s2i_args}")

        # Call the migrated function
        result = ctl.s2i_multistage_build(
            app_path=app_path,
            src_image=src_image,
            sec_image=sec_image,
            dst_image=dst_image,
            s2i_args=s2i_args
        )

        if result:
            print("✅ s2i_multistage_build completed successfully!")
            print(f"Built image: {dst_image}")
        else:
            print("❌ s2i_multistage_build failed!")

        return result

    except Exception as e:
        print(f"❌ Test failed with exception: {e}")
        return False

    finally:
        # Clean up test directory
        if test_app_dir.exists():
            shutil.rmtree(test_app_dir)

        # Clean up container test library resources
        ctl.cleanup()


def main():
    """Main function to run the test."""
    print("=" * 60)
    print("Testing migrated ct_s2i_multistage_build function")
    print("=" * 60)

    success = test_s2i_multistage_build()

    print("=" * 60)
    if success:
        print("🎉 Migration test completed successfully!")
    else:
        print("💥 Migration test failed!")
    print("=" * 60)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Example usage of the Container Test Library Python migration.

This demonstrates how to use the migrated Python functionality
that replaces the container-test-lib.sh bash script.
"""

import sys
from pathlib import Path

from container_ci_suite.container_lib import ContainerTestLib

# Add the container_ci_suite to the path
sys.path.insert(0, str(Path(__file__).parent))


def main():
    """Example usage of the Container Test Library."""

    # Initialize the test library
    ct = ContainerTestLib()

    # Set the image name to test
    ct.image_name = "registry.access.redhat.com/ubi8/python-38"

    print("=== Container Test Library Python Migration Example ===")
    print(f"Testing image: {ct.image_name}")

    # Example 1: Pull an image
    print("\n1. Pulling image...")
    if ct.pull_image(ct.image_name):
        print("✓ Image pulled successfully")
    else:
        print("✗ Failed to pull image")
        return 1

    # Example 2: Check if image exists
    print("\n2. Checking if image exists...")
    if ct.is_container_exists(ct.image_name):
        print("✓ Image exists")
    else:
        print("✗ Image does not exist")

    # Example 3: Create and test a container
    print("\n3. Creating a test container...")
    if ct.create_container("test_container", "sleep 30"):
        print("✓ Container created successfully")

        # Get container IP
        try:
            ip = ct.get_cip("test_container")
            print(f"✓ Container IP: {ip}")
        except Exception as e:
            print(f"✗ Could not get container IP: {e}")

        # Clean up
        ct.cleanup()
        print("✓ Cleanup completed")
    else:
        print("✗ Failed to create container")

    # Example 4: Test environment variables
    print("\n4. Testing environment variable checking...")
    try:
        # This would normally be used with a running container
        print("✓ Environment variable checking functionality available")
    except Exception as e:
        print(f"✗ Environment variable check failed: {e}")

    # Example 5: Show system resources
    print("\n5. Showing system resources...")
    ct.show_resources()

    # Example 6: Generate random string
    print("\n6. Generating random string...")
    random_str = ct.random_string(8)
    print(f"✓ Generated random string: {random_str}")

    # Example 7: Test command assertions
    print("\n7. Testing command assertions...")
    if ct.assert_cmd_success("echo", "test"):
        print("✓ Command assertion (success) works")
    else:
        print("✗ Command assertion (success) failed")

    if ct.assert_cmd_failure("false"):
        print("✓ Command assertion (failure) works")
    else:
        print("✗ Command assertion (failure) failed")

    # Example 8: S2I Multistage Build (new migrated function)
    print("\n8. Testing S2I Multistage Build...")
    try:
        # This is a demonstration - would need real app path and images
        print("✓ S2I Multistage Build function available")
        print("  Usage: ct.s2i_multistage_build(app_path, src_image, sec_image, dst_image, s2i_args)")
        print("  This function creates a multistage Docker build for S2I applications")
        print("  - First stage: builds the application using the source image")
        print("  - Second stage: copies artifacts to a minimal runtime image")
    except Exception as e:
        print(f"✗ S2I Multistage Build test failed: {e}")

    # Example 9: Test App Dockerfile (new migrated function)
    print("\n9. Testing App Dockerfile...")
    try:
        # This is a demonstration - would need real dockerfile and app
        print("✓ Test App Dockerfile function available")
        print("  Usage: ct.test_app_dockerfile(dockerfile, app_url, expected_text, app_dir, build_args)")
        print("  This function:")
        print("  - Builds a Docker image from a Dockerfile and application source")
        print("  - Runs the container and tests HTTP responses")
        print("  - Validates that the application responds with expected content")
        print("  - Handles both local directories and git repositories as app sources")
    except Exception as e:
        print(f"✗ Test App Dockerfile test failed: {e}")

    print("\n=== Example completed successfully ===")
    return 0


if __name__ == "__main__":
    sys.exit(main())



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os
import subprocess

from pathlib import Path
from git import Repo

from container_ci_suite.utils import ContainerTestLibUtils
from container_ci_suite.utils import cwd


logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
logger = logging.getLogger(__name__)


class Git:

    def __init__(self, path: Path, create_dir: bool = False):
        self.path: Path = path
        self._repo = None
        if create_dir:
            logger.info(f"Creating git repo in path {path}")
            self.path.mkdir()

    @property
    def repo(self):
        if not self._repo:
            self._repo = Repo.init(self.path)
            assert not self._repo.bare
        return self._repo

    def add_files(self):
        """
        Function adds all files to git
        """
        print(f"WOrking dir {os.getcwd()}")
        self.repo.git.add(all=True)
        # ContainerTestLibUtils.run_command("git add -A")

    def add_global_config(self, username: str, mail: str):
        """
        Function adds username and mail to git config file
        :param username: str, Specify username added to git config file
        :param mail: str, Specify mail added to git config file
        """
        logger.info(f"Adding to repo username {username} and email {mail}")
        self.repo.config_writer().set_value("user", "name", username).release()
        self.repo.config_writer().set_value("user", "email", mail).release()
        # ContainerTestLibUtils.run_command("git config user.email 'build@localhost'")
        # ContainerTestLibUtils.run_command("git config user.name 'builder'")

    def commit_files(self, commit_command: str = "-am", message: str = "init commit"):
        """
        Commit files to repo defined by path
        :param commit_command: str, Default commit command is -am
        :param message: str, Commit message. Default is "init commit"
        """
        logger.info(f"Commit changes to git repo by commit command message {message}")
        self.repo.git.commit(commit_command, message)
        # ContainerTestLibUtils.run_command("git commit -m 'Sample commit'")

    def create_repo(
            self,
            commit_command: str = "Sample commit",
            message: str = "init commit",
            username: str = "build@localhost",
            mail: str = "builder"
    ) -> bool:
        """
        Function adds files into repository, updates global configuration file with username and mail
        and commit changes to the repository
        :param commit_command: str, define commit_command. Default is '-am'
        :param message: str, define commit message. Default is 'init commit'
        :param username: str, define username which will be added to git config
        :param mail: str, define mail which will be added to git config
        """
        git_init: bool = True
        with cwd(self.path):
            try:
                ContainerTestLibUtils.run_command("git init")
                if username and mail:
                    self.add_global_config(username=username, mail=mail)
                self.add_files()
                self.commit_files(commit_command, message=message)
            except subprocess.CalledProcessError:
                print("Git repo initialization FAILED.")
                git_init = False
        return git_init



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import json
import shutil

import logging
import time
import requests
import subprocess

from typing import Dict, List, Any
from pathlib import Path

import urllib3

import container_ci_suite.utils as utils

from container_ci_suite.utils import ContainerTestLibUtils
from container_ci_suite.openshift import OpenShiftAPI
from container_ci_suite.engines.openshift import OpenShiftOperations

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
logger = logging.getLogger(__name__)


class HelmChartsAPI:

    def __init__(
            self,
            path: Path,
            package_name: str,
            tarball_dir: Path,
            delete_prj: bool = True,
            shared_cluster: bool = False,
    ):
        self.path: Path = path
        self.version: str = ""
        self.package_name: str = package_name
        self.tarball_dir = tarball_dir
        self.delete_prj: bool = delete_prj
        if not shared_cluster:
            self.shared_cluster = shared_cluster
        else:
            self.shared_cluster = utils.is_shared_cluster(test_type="helm")
        self.create_prj: bool = True
        self.oc_api = OpenShiftAPI(
            create_prj=self.create_prj, delete_prj=self.delete_prj, shared_cluster=self.shared_cluster, test_type="helm"
        )
        self.pod_json_data: dict = {}
        self.pod_name_prefix: str = ""
        self.namespace = self.set_namespace()
        self.cloned_dir = ""

    @staticmethod
    def run_helm_command(
        cmd, json_output: bool = True, return_output: bool = True, ignore_error: bool = False, shell: bool = True
    ):
        """
        Run docker command:
        """
        json_cmd = "-o json" if json_output else ""
        print(f"run_helm_command: helm {cmd} {json_cmd}")

        return ContainerTestLibUtils.run_command(
            f"helm {cmd} {json_cmd}",
            return_output=return_output,
            ignore_error=ignore_error,
            shell=shell,
        )

    def set_namespace(self):
        return self.oc_api.namespace

    def delete_project(self):
        self.oc_api.delete_project()
        if self.cloned_dir != "" and Path(self.cloned_dir).exists():
            shutil.rmtree(self.cloned_dir)

    @property
    def full_package_dir(self) -> Path:
        return Path(self.path) / self.package_name / "src"

    @property
    def get_tarball_name(self):
        return f"{self.package_name}-{self.version}.tgz"

    @property
    def get_full_tarball_path(self):
        return self.tarball_dir / self.get_tarball_name

    def clone_helm_chart_repo(self, repo_url: str, repo_name: str, subdir: str = ""):
        temp_dir = utils.temporary_dir()
        self.cloned_dir = temp_dir
        cmd_clone = f"git clone {repo_url} {temp_dir}/{repo_name}"
        print(f"Clone charts repo by command: {cmd_clone}")
        clone_output = ContainerTestLibUtils.run_command(cmd_clone, return_output=True)
        print(clone_output)
        if subdir != "":
            self.path = Path(temp_dir) / repo_name / subdir
        else:
            self.path = Path(temp_dir) / repo_name

    def get_version_from_chart_yaml(self) -> Any:
        chart_dict = utils.get_yaml_data(self.full_package_dir / "Chart.yaml")
        if "appVersion" in chart_dict:
            return chart_dict["appVersion"]
        return None

    def is_registry_in_values_yaml(self) -> bool:
        chart_dict = utils.get_yaml_data(self.full_package_dir / "values.yaml")
        if "registry" in chart_dict:
            return True
        return False

    def is_pvc_in_values_yaml(self) -> bool:
        chart_dict = utils.get_yaml_data(self.full_package_dir / "values.yaml")
        if "pvc" in chart_dict:
            return True
        return False

    def get_name_from_values_yaml(self) -> Any:
        chart_dict = utils.get_yaml_data(self.full_package_dir / "values.yaml")
        if "name" in chart_dict:
            return chart_dict["name"]
        return None

    def set_version(self, version: str):
        self.version = version

    def is_chart_yaml_present(self):
        if (self.full_package_dir / "Chart.yaml") .exists():
            return True
        return False

    def is_s2i_pod_running(self, pod_name_prefix: str, timeout: int = 180):
        oc_ops = OpenShiftOperations()
        oc_ops.set_namespace(namespace=self.oc_api.namespace)
        return oc_ops.is_s2i_pod_running(pod_name_prefix=pod_name_prefix, cycle_count=timeout)

    def is_pod_running(self, pod_name_prefix: str, loops: int = 180):
        oc_ops = OpenShiftOperations()
        oc_ops.set_namespace(namespace=self.oc_api.namespace)
        return oc_ops.is_pod_running(pod_name_prefix=pod_name_prefix, loops=loops)

    def helm_package(self) -> bool:
        """
        Package source to Helm Chart package
        """
        if not self.is_chart_yaml_present():
            print(f"Chart.yaml file is not present in directory {self.full_package_dir}")
            return False
        self.version = self.get_version_from_chart_yaml()
        print(f"Helm package command is: helm package {self.full_package_dir}")
        output = HelmChartsAPI.run_helm_command(f"package {self.full_package_dir}", json_output=False)
        print(output)
        if "Successfully packaged chart" in output:
            print(self.get_tarball_name)
            if self.get_tarball_name in output:
                return True
        return False

    def get_helm_json_output(self, command: str) -> Dict:
        try:
            output = HelmChartsAPI.run_helm_command(cmd=command)
        except subprocess.CalledProcessError as cpe:
            print(f"Helm command {command} failed. See {cpe.output}")
            return {}
        # Remove debug wrong output
        new_output = []
        for line in output.split('\n'):
            if line.startswith("W"):
                continue
            new_output.append(line)
        # output = [x for x in output.split('\n') if not x.startswith("W")]
        # print(output)
        return json.loads(''.join(new_output))

    def check_helm_installation(self):
        # Let's check that pod is really running
        json_output = self.get_helm_json_output(command="list")
        for out in json_output:
            if out["name"] != self.package_name:
                continue
            assert out["name"] == self.package_name
            assert out["chart"] == f"{self.package_name}-{self.version}"
            assert out["status"] == "deployed"
            return True
        return False

    def is_helm_package_installed(self):
        json_output = self.get_helm_json_output(command="list")
        for out in json_output:
            if out["name"] == self.package_name:
                return True
        return False

    def helm_uninstallation(self):
        output = HelmChartsAPI.run_helm_command(
            f"uninstall {self.package_name} -n {self.oc_api.namespace}", json_output=False
        )
        print(output)

    def helm_installation(self, values: Dict = None):
        self.version = self.get_version_from_chart_yaml()
        if not self.version:
            return False
        if self.is_helm_package_installed():
            self.helm_uninstallation()
        command_values = ""
        if values:
            command_values += " " + ' '.join([f"--set {key}={value}" for key, value in values.items()])
        if self.shared_cluster:
            if self.is_registry_in_values_yaml():
                if self.shared_cluster:
                    command_values += " " + ' '.join(
                        [f"--set {key}={value}" for key, value in utils.shared_cluster_variables().items()]
                    )
            if self.is_pvc_in_values_yaml():
                command_values += (f" --set pvc.netapp_nfs=true "
                                   f"--set pvc.app_code={utils.get_shared_variable('app_code')}")
        install_success: bool = False
        json_output: Dict = {}
        for count in range(3):
            json_output = self.get_helm_json_output(
                f"install {self.package_name} {self.get_full_tarball_path} {command_values}"
            )
            if json_output:
                install_success = True
                break
            time.sleep(3)
        # Let's wait couple seconds, till it is not really imported
        time.sleep(3)
        if not install_success:
            return False
        assert json_output["name"] == self.package_name
        assert json_output["chart"]["metadata"]["version"] == self.version
        assert json_output["info"]["status"] == "deployed"
        if not self.check_helm_installation():
            print("Installation has failed. Let's uninstall it and try one more time.")
            return False
        return True

    def check_test_output(self, output, expected_str: List[str]):
        """
        Expected output from helm test is e.g.:
        NAME: postgresql-persistent
        LAST DEPLOYED: Wed Apr 12 08:08:20 2023
        NAMESPACE: helm-sclorg-7072
        STATUS: deployed
        REVISION: 1
        TEST SUITE:     postgresql-persistent-connection-test
        Last Started:   Wed Apr 12 08:08:52 2023
        Last Completed: Wed Apr 12 08:08:56 2023
        Phase:          Succeeded

        POD LOGS: postgresql-persistent-connection-test
        postgresql-testing:5432 - accepting connections
        """
        check_list = [
            f"NAME: {self.package_name}",
            "STATUS: deployed",
            "Succeeded"
        ]
        if isinstance(expected_str, str):
            print("Function expects list of strings to check.")
            return False
        check_list.extend(expected_str)
        print(f"Strings to check in helm output log: {check_list}")
        result_list = [x in ''.join(output) for x in check_list]
        if False in result_list:
            return False
        return True

    def test_helm_chart(self, expected_str: List[str]) -> bool:
        for count in range(60):
            time.sleep(2)
            try:
                output = HelmChartsAPI.run_helm_command(
                    f"test {self.package_name} --logs", json_output=False
                )
                print(f"Helm test output: {output}")
            except subprocess.CalledProcessError:
                print(f"Helm test command `test {self.package_name} --logs` failed. Let's try more time.")
                continue
            if self.check_test_output(output, expected_str=expected_str):
                return True
        oc_ops = OpenShiftOperations()
        oc_ops.set_namespace(namespace=self.oc_api.namespace)
        oc_ops.print_get_status()
        return False

    def get_route_name(self, route_name: str):
        oc_ops = OpenShiftOperations()
        oc_ops.set_namespace(namespace=self.oc_api.namespace)
        json_data = oc_ops.get_routes()
        if len(json_data["items"]) == 0:
            return None
        for item in json_data["items"]:
            if item["metadata"]["namespace"] != self.oc_api.namespace:
                continue
            if item["metadata"]["name"] != route_name:
                continue
            if item["spec"]["to"]["name"] != route_name:
                continue
            return item["spec"]["host"]
        return None

    def check_imagestreams(self, version: str, registry: str) -> bool:
        """
        Returns JSON output
        """
        oc_ops = OpenShiftOperations()
        oc_ops.set_namespace(namespace=self.oc_api.namespace)
        json_output = oc_ops.oc_gel_all_is()
        for tag in json_output["items"][0]["spec"]["tags"]:
            tag_name = tag["name"]
            tag_registry = tag["from"]["name"]
            print(f"Important tags: {version}={tag_name}, {registry}={tag_registry}")
            if tag_name == version and tag_registry == registry:
                print("Imagestream tag exists.")
                return True
        return False

    def test_helm_curl_output(
            self, route_name: str, expected_str: str, port: int = None, schema: str = "http://",
            range_count: int = 10
    ) -> bool:
        # Let's get some time to start application
        time.sleep(3)
        host_name = self.get_route_name(route_name=route_name)
        print(f"test_helm_curl_output: Route name is: {host_name}")
        if not host_name:
            return False
        url_address = f"{schema}{host_name}"
        if port:
            url_address = f"{url_address}:{port}"
        valid_request: bool = False
        for count in range(range_count):
            try:
                print(f"test_helm_curl_output: requests.get {url_address}")
                resp = requests.get(url_address, verify=False)
                resp.raise_for_status()
                if resp.status_code != 200:
                    print(f"test_helm_curl_output response is different from 200: {resp.text}, {resp.status_code}")
                    continue
                print(f"test_helm_curl_output: text: {resp.text}")
                if expected_str not in resp.text:
                    print(f"{expected_str} is not in the output")
                    continue
                return True
            except requests.exceptions.HTTPError:
                print("test_helm_curl_output: Service is not yet available. Let's wait some time")
                time.sleep(3)
                continue
            except requests.exceptions.ConnectTimeout:
                print("test_helm_curl_output: Service is not yet available. Connection Timeout")
                time.sleep(3)
                continue
            except urllib3.exceptions.MaxRetryError:
                print("test_helm_curl_output: MaxRetryError. Let's wait some time")
                time.sleep(3)
                continue
        if not valid_request:
            print("test_helm_curl_output: Service was not available")
            return False
        return False
